// Copyright (c) 2015 Aaron Power
// Use of this source code is governed by the APACHE2.0/MIT licence that can be
// found in the LICENCE-{APACHE/MIT} file.

use std::borrow::Cow;
use std::fmt;
use std::path::Path;
use std::fs::File;
use std::io::{BufRead, BufReader};

use utils::fs as fsutils;
use self::LanguageType::*;
use Languages;
use Language;


#[cfg_attr(feature = "io", derive(Deserialize, Serialize))]
#[derive(Clone, Copy, Debug, Eq, Ord, PartialEq, PartialOrd)]
pub enum LanguageType {Abap,
    ActionScript,
    Ada,
    Agda,
    Alex,
    Asp,
    AspNet,
    Assembly,
    AutoHotKey,
    Autoconf,
    Bash,
    Batch,
    C,
    CHeader,
    CMake,
    CSharp,
    CShell,
    Cabal,
    Cassius,
    Ceylon,
    Clojure,
    ClojureC,
    ClojureScript,
    Cobol,
    CoffeeScript,
    Cogent,
    ColdFusion,
    ColdFusionScript,
    Coq,
    Cpp,
    CppHeader,
    Crystal,
    Css,
    D,
    Dart,
    DeviceTree,
    Dockerfile,
    Elisp,
    Elixir,
    Elm,
    Elvish,
    EmacsDevEnv,
    Erlang,
    FEN,
    FSharp,
    Fish,
    Forth,
    FortranLegacy,
    FortranModern,
    Fstar,
    GdScript,
    Glsl,
    Go,
    Groovy,
    Hamlet,
    Handlebars,
    Happy,
    Haskell,
    Haxe,
    Hcl,
    Hex,
    Html,
    Idris,
    IntelHex,
    Isabelle,
    Jai,
    Java,
    JavaScript,
    Json,
    Jsx,
    Julia,
    Julius,
    KakouneScript,
    Kotlin,
    Lean,
    Less,
    LinkerScript,
    Lisp,
    Lua,
    Lucius,
    Madlang,
    Makefile,
    Markdown,
    Meson,
    Mint,
    ModuleDef,
    MsBuild,
    Mustache,
    Nim,
    Nix,
    OCaml,
    ObjectiveC,
    ObjectiveCpp,
    Org,
    Oz,
    PSL,
    Pascal,
    Perl,
    Php,
    Polly,
    Processing,
    Prolog,
    Protobuf,
    PureScript,
    Python,
    Qcl,
    Qml,
    R,
    Racket,
    Rakefile,
    Razor,
    ReStructuredText,
    Ruby,
    RubyHtml,
    Rust,
    SRecode,
    Sass,
    Scala,
    Scheme,
    Scons,
    Sh,
    Sml,
    SpecmanE,
    Spice,
    Sql,
    Svg,
    Swift,
    SystemVerilog,
    Tcl,
    Tex,
    Text,
    Toml,
    TypeScript,
    UnrealScript,
    UrWeb,
    UrWebProject,
    Vala,
    Verilog,
    VerilogArgsFile,
    Vhdl,
    VimScript,
    VisualBasic,
    Vue,
    Wolfram,
    XSL,
    Xaml,
    Xml,
    Xtend,
    Yaml,
    Zig,
    Zsh,
    
}

impl LanguageType {
    /// Returns the display name of a language.
    ///
    /// ```
    /// # use tokei::*;
    /// let bash = LanguageType::Bash;
    ///
    /// assert_eq!(bash.name(), "BASH");
    /// ```
    pub fn name(&self) -> &'static str {
        match *self {
                Abap =>
                
                    "ABAP"
                ,
                ActionScript =>
                
                    "ActionScript",
                Ada =>
                
                    "Ada",
                Agda =>
                
                    "Agda",
                Alex =>
                
                    "Alex",
                Asp =>
                
                    "ASP"
                ,
                AspNet =>
                
                    "ASP.NET"
                ,
                Assembly =>
                
                    "Assembly",
                AutoHotKey =>
                
                    "AutoHotKey",
                Autoconf =>
                
                    "Autoconf",
                Bash =>
                
                    "BASH"
                ,
                Batch =>
                
                    "Batch",
                C =>
                
                    "C",
                CHeader =>
                
                    "C Header"
                ,
                CMake =>
                
                    "CMake",
                CSharp =>
                
                    "C#"
                ,
                CShell =>
                
                    "C Shell"
                ,
                Cabal =>
                
                    "Cabal",
                Cassius =>
                
                    "Cassius",
                Ceylon =>
                
                    "Ceylon",
                Clojure =>
                
                    "Clojure",
                ClojureC =>
                
                    "ClojureC",
                ClojureScript =>
                
                    "ClojureScript",
                Cobol =>
                
                    "COBOL"
                ,
                CoffeeScript =>
                
                    "CoffeeScript",
                Cogent =>
                
                    "Cogent",
                ColdFusion =>
                
                    "ColdFusion",
                ColdFusionScript =>
                
                    "ColdFusion CFScript"
                ,
                Coq =>
                
                    "Coq",
                Cpp =>
                
                    "C++"
                ,
                CppHeader =>
                
                    "C++ Header"
                ,
                Crystal =>
                
                    "Crystal",
                Css =>
                
                    "CSS"
                ,
                D =>
                
                    "D",
                Dart =>
                
                    "Dart",
                DeviceTree =>
                
                    "Device Tree"
                ,
                Dockerfile =>
                
                    "Dockerfile",
                Elisp =>
                
                    "Emacs Lisp"
                ,
                Elixir =>
                
                    "Elixir",
                Elm =>
                
                    "Elm",
                Elvish =>
                
                    "Elvish",
                EmacsDevEnv =>
                
                    "Emacs Dev Env"
                ,
                Erlang =>
                
                    "Erlang",
                FEN =>
                
                    "FEN"
                ,
                FSharp =>
                
                    "F#"
                ,
                Fish =>
                
                    "Fish",
                Forth =>
                
                    "Forth",
                FortranLegacy =>
                
                    "FORTRAN Legacy"
                ,
                FortranModern =>
                
                    "FORTRAN Modern"
                ,
                Fstar =>
                
                    "F*"
                ,
                GdScript =>
                
                    "GDScript"
                ,
                Glsl =>
                
                    "GLSL"
                ,
                Go =>
                
                    "Go",
                Groovy =>
                
                    "Groovy",
                Hamlet =>
                
                    "Hamlet",
                Handlebars =>
                
                    "Handlebars",
                Happy =>
                
                    "Happy",
                Haskell =>
                
                    "Haskell",
                Haxe =>
                
                    "Haxe",
                Hcl =>
                
                    "HCL"
                ,
                Hex =>
                
                    "HEX"
                ,
                Html =>
                
                    "HTML"
                ,
                Idris =>
                
                    "Idris",
                IntelHex =>
                
                    "Intel HEX"
                ,
                Isabelle =>
                
                    "Isabelle",
                Jai =>
                
                    "JAI"
                ,
                Java =>
                
                    "Java",
                JavaScript =>
                
                    "JavaScript",
                Json =>
                
                    "JSON"
                ,
                Jsx =>
                
                    "JSX"
                ,
                Julia =>
                
                    "Julia",
                Julius =>
                
                    "Julius",
                KakouneScript =>
                
                    "Kakoune script"
                ,
                Kotlin =>
                
                    "Kotlin",
                Lean =>
                
                    "Lean",
                Less =>
                
                    "LESS"
                ,
                LinkerScript =>
                
                    "LD Script"
                ,
                Lisp =>
                
                    "Lisp",
                Lua =>
                
                    "Lua",
                Lucius =>
                
                    "Lucius",
                Madlang =>
                
                    "Madlang",
                Makefile =>
                
                    "Makefile",
                Markdown =>
                
                    "Markdown",
                Meson =>
                
                    "Meson",
                Mint =>
                
                    "Mint",
                ModuleDef =>
                
                    "Module-Definition"
                ,
                MsBuild =>
                
                    "MSBuild"
                ,
                Mustache =>
                
                    "Mustache",
                Nim =>
                
                    "Nim",
                Nix =>
                
                    "Nix",
                OCaml =>
                
                    "OCaml",
                ObjectiveC =>
                
                    "Objective C"
                ,
                ObjectiveCpp =>
                
                    "Objective C++"
                ,
                Org =>
                
                    "Org",
                Oz =>
                
                    "Oz",
                PSL =>
                
                    "PSL Assertion"
                ,
                Pascal =>
                
                    "Pascal",
                Perl =>
                
                    "Perl",
                Php =>
                
                    "PHP"
                ,
                Polly =>
                
                    "Polly",
                Processing =>
                
                    "Processing",
                Prolog =>
                
                    "Prolog",
                Protobuf =>
                
                    "Protocol Buffers"
                ,
                PureScript =>
                
                    "PureScript",
                Python =>
                
                    "Python",
                Qcl =>
                
                    "QCL"
                ,
                Qml =>
                
                    "QML"
                ,
                R =>
                
                    "R",
                Racket =>
                
                    "Racket",
                Rakefile =>
                
                    "Rakefile",
                Razor =>
                
                    "Razor",
                ReStructuredText =>
                
                    "ReStructuredText",
                Ruby =>
                
                    "Ruby",
                RubyHtml =>
                
                    "Ruby HTML"
                ,
                Rust =>
                
                    "Rust",
                SRecode =>
                
                    "SRecode Template"
                ,
                Sass =>
                
                    "Sass",
                Scala =>
                
                    "Scala",
                Scheme =>
                
                    "Scheme",
                Scons =>
                
                    "Scons",
                Sh =>
                
                    "Shell"
                ,
                Sml =>
                
                    "Standard ML (SML)"
                ,
                SpecmanE =>
                
                    "Specman e"
                ,
                Spice =>
                
                    "Spice Netlist"
                ,
                Sql =>
                
                    "SQL"
                ,
                Svg =>
                
                    "SVG"
                ,
                Swift =>
                
                    "Swift",
                SystemVerilog =>
                
                    "SystemVerilog",
                Tcl =>
                
                    "TCL"
                ,
                Tex =>
                
                    "TeX"
                ,
                Text =>
                
                    "Plain Text"
                ,
                Toml =>
                
                    "TOML"
                ,
                TypeScript =>
                
                    "TypeScript",
                UnrealScript =>
                
                    "Unreal Script"
                ,
                UrWeb =>
                
                    "Ur/Web"
                ,
                UrWebProject =>
                
                    "Ur/Web Project"
                ,
                Vala =>
                
                    "Vala",
                Verilog =>
                
                    "Verilog",
                VerilogArgsFile =>
                
                    "Verilog Args File"
                ,
                Vhdl =>
                
                    "VHDL"
                ,
                VimScript =>
                
                    "Vim Script"
                ,
                VisualBasic =>
                
                    "Visual Basic"
                ,
                Vue =>
                
                    "Vue"
                ,
                Wolfram =>
                
                    "Wolfram",
                XSL =>
                
                    "XSL"
                ,
                Xaml =>
                
                    "XAML"
                ,
                Xml =>
                
                    "XML"
                ,
                Xtend =>
                
                    "Xtend",
                Yaml =>
                
                    "YAML"
                ,
                Zig =>
                
                    "Zig",
                Zsh =>
                
                    "Zsh",
        }
    }

    /// Provides every variant in a Vec
    pub fn list() -> Vec<Self> {
        return vec! [
            
                Abap,
                ActionScript,
                Ada,
                Agda,
                Alex,
                Asp,
                AspNet,
                Assembly,
                AutoHotKey,
                Autoconf,
                Bash,
                Batch,
                C,
                CHeader,
                CMake,
                CSharp,
                CShell,
                Cabal,
                Cassius,
                Ceylon,
                Clojure,
                ClojureC,
                ClojureScript,
                Cobol,
                CoffeeScript,
                Cogent,
                ColdFusion,
                ColdFusionScript,
                Coq,
                Cpp,
                CppHeader,
                Crystal,
                Css,
                D,
                Dart,
                DeviceTree,
                Dockerfile,
                Elisp,
                Elixir,
                Elm,
                Elvish,
                EmacsDevEnv,
                Erlang,
                FEN,
                FSharp,
                Fish,
                Forth,
                FortranLegacy,
                FortranModern,
                Fstar,
                GdScript,
                Glsl,
                Go,
                Groovy,
                Hamlet,
                Handlebars,
                Happy,
                Haskell,
                Haxe,
                Hcl,
                Hex,
                Html,
                Idris,
                IntelHex,
                Isabelle,
                Jai,
                Java,
                JavaScript,
                Json,
                Jsx,
                Julia,
                Julius,
                KakouneScript,
                Kotlin,
                Lean,
                Less,
                LinkerScript,
                Lisp,
                Lua,
                Lucius,
                Madlang,
                Makefile,
                Markdown,
                Meson,
                Mint,
                ModuleDef,
                MsBuild,
                Mustache,
                Nim,
                Nix,
                OCaml,
                ObjectiveC,
                ObjectiveCpp,
                Org,
                Oz,
                PSL,
                Pascal,
                Perl,
                Php,
                Polly,
                Processing,
                Prolog,
                Protobuf,
                PureScript,
                Python,
                Qcl,
                Qml,
                R,
                Racket,
                Rakefile,
                Razor,
                ReStructuredText,
                Ruby,
                RubyHtml,
                Rust,
                SRecode,
                Sass,
                Scala,
                Scheme,
                Scons,
                Sh,
                Sml,
                SpecmanE,
                Spice,
                Sql,
                Svg,
                Swift,
                SystemVerilog,
                Tcl,
                Tex,
                Text,
                Toml,
                TypeScript,
                UnrealScript,
                UrWeb,
                UrWebProject,
                Vala,
                Verilog,
                VerilogArgsFile,
                Vhdl,
                VimScript,
                VisualBasic,
                Vue,
                Wolfram,
                XSL,
                Xaml,
                Xml,
                Xtend,
                Yaml,
                Zig,
                Zsh,
        ]
    }

    /// Get language from a file path. May open and read the file.
    ///
    /// ```no_run
    /// # use tokei::*;
    /// let rust = LanguageType::from_path("./main.rs");
    ///
    /// assert_eq!(rust, Some(LanguageType::Rust));
    /// ```
    pub fn from_path<P: AsRef<Path>>(entry: P) -> Option<Self> {

        if let Some(filename) = fsutils::get_filename(&entry) {
            match &*filename {
                            "cmakelists.txt"
                            => return Some(CMake),
                            "dockerfile"
                            => return Some(Dockerfile),
                            "makefile"
                            => return Some(Makefile),
                            "meson.build" |
                            "meson_options.txt"
                            => return Some(Meson),
                            "rakefile"
                            => return Some(Rakefile),
                            "sconstruct" |
                            "sconscript"
                            => return Some(Scons),
                _ => ()
            }
        }

        let extension = fsutils::get_extension(&entry)
            .or_else(|| get_filetype_from_shebang(&entry).map(String::from));

        if let Some(extension) = extension {
            match &*extension {
                            "abap"
                            => Some(Abap),
                            "as"
                            => Some(ActionScript),
                            "ada" |
                            "adb" |
                            "ads" |
                            "pad"
                            => Some(Ada),
                            "agda"
                            => Some(Agda),
                            "x"
                            => Some(Alex),
                            "asa" |
                            "asp"
                            => Some(Asp),
                            "asax" |
                            "ascx" |
                            "asmx" |
                            "aspx" |
                            "master" |
                            "sitemap" |
                            "webinfo"
                            => Some(AspNet),
                            "s" |
                            "asm"
                            => Some(Assembly),
                            "ahk"
                            => Some(AutoHotKey),
                            "in"
                            => Some(Autoconf),
                            "bash"
                            => Some(Bash),
                            "bat" |
                            "btm" |
                            "cmd"
                            => Some(Batch),
                            "c" |
                            "ec" |
                            "pgc"
                            => Some(C),
                            "h"
                            => Some(CHeader),
                            "cmake"
                            => Some(CMake),
                            "cs"
                            => Some(CSharp),
                            "csh"
                            => Some(CShell),
                            "cabal"
                            => Some(Cabal),
                            "cassius"
                            => Some(Cassius),
                            "ceylon"
                            => Some(Ceylon),
                            "clj"
                            => Some(Clojure),
                            "cljc"
                            => Some(ClojureC),
                            "cljs"
                            => Some(ClojureScript),
                            "cob" |
                            "cbl" |
                            "ccp" |
                            "cobol" |
                            "cpy"
                            => Some(Cobol),
                            "coffee"
                            => Some(CoffeeScript),
                            "cogent"
                            => Some(Cogent),
                            "cfm"
                            => Some(ColdFusion),
                            "cfc"
                            => Some(ColdFusionScript),
                            "v"
                            => Some(Coq),
                            "cc" |
                            "cpp" |
                            "cxx" |
                            "c++" |
                            "pcc"
                            => Some(Cpp),
                            "hh" |
                            "hpp" |
                            "hxx" |
                            "inl" |
                            "ipp"
                            => Some(CppHeader),
                            "cr"
                            => Some(Crystal),
                            "css"
                            => Some(Css),
                            "d"
                            => Some(D),
                            "dart"
                            => Some(Dart),
                            "dts" |
                            "dtsi"
                            => Some(DeviceTree),
                            "dockerfile" |
                            "dockerignore"
                            => Some(Dockerfile),
                            "el"
                            => Some(Elisp),
                            "ex" |
                            "exs"
                            => Some(Elixir),
                            "elm"
                            => Some(Elm),
                            "elv"
                            => Some(Elvish),
                            "ede"
                            => Some(EmacsDevEnv),
                            "erl" |
                            "hrl"
                            => Some(Erlang),
                            "fen"
                            => Some(FEN),
                            "fs" |
                            "fsi" |
                            "fsx" |
                            "fsscript"
                            => Some(FSharp),
                            "fish"
                            => Some(Fish),
                            "4th" |
                            "forth" |
                            "fr" |
                            "frt" |
                            "fth" |
                            "f83" |
                            "fb" |
                            "fpm" |
                            "e4" |
                            "rx" |
                            "ft"
                            => Some(Forth),
                            "f" |
                            "for" |
                            "ftn" |
                            "f77" |
                            "pfo"
                            => Some(FortranLegacy),
                            "f03" |
                            "f08" |
                            "f90" |
                            "f95"
                            => Some(FortranModern),
                            "fst"
                            => Some(Fstar),
                            "gd"
                            => Some(GdScript),
                            "vert" |
                            "tesc" |
                            "tese" |
                            "geom" |
                            "frag" |
                            "comp"
                            => Some(Glsl),
                            "go"
                            => Some(Go),
                            "groovy" |
                            "grt" |
                            "gtpl" |
                            "gvy"
                            => Some(Groovy),
                            "hamlet"
                            => Some(Hamlet),
                            "hbs" |
                            "handlebars"
                            => Some(Handlebars),
                            "y" |
                            "ly"
                            => Some(Happy),
                            "hs"
                            => Some(Haskell),
                            "hx"
                            => Some(Haxe),
                            "tf" |
                            "tfvars"
                            => Some(Hcl),
                            "hex"
                            => Some(Hex),
                            "html" |
                            "htm"
                            => Some(Html),
                            "idr" |
                            "lidr"
                            => Some(Idris),
                            "ihex"
                            => Some(IntelHex),
                            "thy"
                            => Some(Isabelle),
                            "jai"
                            => Some(Jai),
                            "java"
                            => Some(Java),
                            "js"
                            => Some(JavaScript),
                            "json"
                            => Some(Json),
                            "jsx"
                            => Some(Jsx),
                            "jl"
                            => Some(Julia),
                            "julius"
                            => Some(Julius),
                            "kak"
                            => Some(KakouneScript),
                            "kt" |
                            "kts"
                            => Some(Kotlin),
                            "lean" |
                            "hlean"
                            => Some(Lean),
                            "less"
                            => Some(Less),
                            "lds"
                            => Some(LinkerScript),
                            "lisp" |
                            "lsp"
                            => Some(Lisp),
                            "lua"
                            => Some(Lua),
                            "lucius"
                            => Some(Lucius),
                            "mad"
                            => Some(Madlang),
                            "makefile" |
                            "mak" |
                            "mk"
                            => Some(Makefile),
                            "md" |
                            "markdown"
                            => Some(Markdown),
                            "mint"
                            => Some(Mint),
                            "def"
                            => Some(ModuleDef),
                            "csproj" |
                            "vbproj" |
                            "fsproj" |
                            "props" |
                            "targets"
                            => Some(MsBuild),
                            "mustache"
                            => Some(Mustache),
                            "nim"
                            => Some(Nim),
                            "nix"
                            => Some(Nix),
                            "ml" |
                            "mli" |
                            "re" |
                            "rei"
                            => Some(OCaml),
                            "m"
                            => Some(ObjectiveC),
                            "mm"
                            => Some(ObjectiveCpp),
                            "org"
                            => Some(Org),
                            "oz"
                            => Some(Oz),
                            "psl"
                            => Some(PSL),
                            "pas"
                            => Some(Pascal),
                            "pl" |
                            "pm"
                            => Some(Perl),
                            "php"
                            => Some(Php),
                            "polly"
                            => Some(Polly),
                            "pde"
                            => Some(Processing),
                            "p" |
                            "pro"
                            => Some(Prolog),
                            "proto"
                            => Some(Protobuf),
                            "purs"
                            => Some(PureScript),
                            "py"
                            => Some(Python),
                            "qcl"
                            => Some(Qcl),
                            "qml"
                            => Some(Qml),
                            "r"
                            => Some(R),
                            "rkt"
                            => Some(Racket),
                            "rake"
                            => Some(Rakefile),
                            "cshtml"
                            => Some(Razor),
                            "rst"
                            => Some(ReStructuredText),
                            "rb"
                            => Some(Ruby),
                            "rhtml"
                            => Some(RubyHtml),
                            "rs"
                            => Some(Rust),
                            "srt"
                            => Some(SRecode),
                            "sass" |
                            "scss"
                            => Some(Sass),
                            "sc" |
                            "scala"
                            => Some(Scala),
                            "scm" |
                            "ss"
                            => Some(Scheme),
                            "sh"
                            => Some(Sh),
                            "sml"
                            => Some(Sml),
                            "e"
                            => Some(SpecmanE),
                            "ckt"
                            => Some(Spice),
                            "sql"
                            => Some(Sql),
                            "svg"
                            => Some(Svg),
                            "swift"
                            => Some(Swift),
                            "sv" |
                            "svh"
                            => Some(SystemVerilog),
                            "tcl"
                            => Some(Tcl),
                            "tex" |
                            "sty"
                            => Some(Tex),
                            "text" |
                            "txt"
                            => Some(Text),
                            "toml"
                            => Some(Toml),
                            "ts" |
                            "tsx"
                            => Some(TypeScript),
                            "uc" |
                            "uci" |
                            "upkg"
                            => Some(UnrealScript),
                            "ur" |
                            "urs"
                            => Some(UrWeb),
                            "urp"
                            => Some(UrWebProject),
                            "vala"
                            => Some(Vala),
                            "vg" |
                            "vh"
                            => Some(Verilog),
                            "irunargs" |
                            "xrunargs"
                            => Some(VerilogArgsFile),
                            "vhd"
                            => Some(Vhdl),
                            "vim"
                            => Some(VimScript),
                            "vb"
                            => Some(VisualBasic),
                            "vue"
                            => Some(Vue),
                            "nb" |
                            "wl"
                            => Some(Wolfram),
                            "xsl"
                            => Some(XSL),
                            "xaml"
                            => Some(Xaml),
                            "xml"
                            => Some(Xml),
                            "xtend"
                            => Some(Xtend),
                            "yaml" |
                            "yml"
                            => Some(Yaml),
                            "zig"
                            => Some(Zig),
                            "zsh"
                            => Some(Zsh),
                extension => {
                    warn!("Unknown extension: {}", extension);
                    None
                },
            }
        } else {
            None
        }
    }
}

impl Languages {
    pub fn generate_language(language: LanguageType) -> Language {
        match language {
            Abap =>
                            Language::new_single(&[
                                    "*",
                                    "\"",
                            ])
                    ,ActionScript =>
                    Language::new_c()
                ,Ada =>
                            Language::new_single(&[
                                    "--",
                            ])
                    ,Agda =>
                    Language::new_haskell()
                ,Alex =>
                        Language::new_multi(&[
                        ]),Asp =>
                            Language::new_single(&[
                                    "'",
                                    "REM",
                            ])
                    ,AspNet =>
                        Language::new_multi(&[
                                (
                                    "<!--",
                                    "-->",
                                ),
                                (
                                    "<%--",
                                    "-->",
                                ),
                        ]),Assembly =>
                            Language::new_single(&[
                                    ";",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),AutoHotKey =>
                            Language::new(
                                &[
                                    ";",
                                ],
                                &[
                                    (
                                        "/*",
                                        "*/",
                                    ),
                                ]
                            )
                        
                    ,Autoconf =>
                            Language::new_single(&[
                                    "#",
                                    "dnl",
                            ])
                    ,Bash =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Batch =>
                            Language::new_single(&[
                                    "REM",
                                    "::",
                            ])
                    ,C =>
                    Language::new_c()
                ,CHeader =>
                    Language::new_c()
                ,CMake =>
                            Language::new_single(&[
                                    "#",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                    ]),CSharp =>
                    Language::new_c()
                ,CShell =>
                    Language::new_hash()
                ,Cabal =>
                    Language::new_haskell()
                ,Cassius =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Ceylon =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Clojure =>
                            Language::new_single(&[
                                    ";",
                            ])
                    ,ClojureC =>
                            Language::new_single(&[
                                    ";",
                            ])
                    ,ClojureScript =>
                            Language::new_single(&[
                                    ";",
                            ])
                    ,Cobol =>
                            Language::new_single(&[
                                    "*",
                            ])
                    ,CoffeeScript =>
                            Language::new(
                                &[
                                    "#",
                                ],
                                &[
                                    (
                                        "###",
                                        "###",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Cogent =>
                            Language::new_single(&[
                                    "--",
                            ])
                    ,ColdFusion =>
                        Language::new_multi(&[
                                (
                                    "<!---",
                                    "--->",
                                ),
                        ])
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),ColdFusionScript =>
                    Language::new_c()
                ,Coq =>
                    Language::new_func()
                ,Cpp =>
                    Language::new_c()
                ,CppHeader =>
                    Language::new_c()
                ,Crystal =>
                            Language::new_single(&[
                                    "#",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Css =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),D =>
                    Language::new_c()
                
                    .nested_comments(&[
                            (
                                "/+",
                                "+/",
                            ),
                    ])
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Dart =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),DeviceTree =>
                    Language::new_c()
                ,Dockerfile =>
                            Language::new_single(&[
                                    "#",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Elisp =>
                            Language::new_single(&[
                                    ";",
                            ])
                    ,Elixir =>
                            Language::new_single(&[
                                    "#",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Elm =>
                    Language::new_haskell()
                ,Elvish =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),EmacsDevEnv =>
                            Language::new_single(&[
                                    ";",
                            ])
                    ,Erlang =>
                            Language::new_single(&[
                                    "%",
                            ])
                    ,FEN =>
                    Language::new_blank()
                ,FSharp =>
                            Language::new(
                                &[
                                    "//",
                                ],
                                &[
                                    (
                                        "(*",
                                        "*)",
                                    ),
                                ]
                            )
                        
                    ,Fish =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Forth =>
                            Language::new(
                                &[
                                    "\\",
                                ],
                                &[
                                    (
                                        "( ",
                                        ")",
                                    ),
                                ]
                            )
                        
                    ,FortranLegacy =>
                            Language::new_single(&[
                                    "c",
                                    "C",
                                    "!",
                                    "*",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),FortranModern =>
                            Language::new_single(&[
                                    "!",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Fstar =>
                    Language::new_func()
                ,GdScript =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Glsl =>
                    Language::new_c()
                ,Go =>
                    Language::new_c()
                ,Groovy =>
                    Language::new_c()
                ,Hamlet =>
                    Language::new_html()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Handlebars =>
                        Language::new_multi(&[
                                (
                                    "<!--",
                                    "-->",
                                ),
                                (
                                    "{{!",
                                    "}}",
                                ),
                        ])
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Happy =>
                        Language::new_multi(&[
                        ]),Haskell =>
                    Language::new_haskell()
                ,Haxe =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Hcl =>
                            Language::new(
                                &[
                                    "#",
                                    "//",
                                ],
                                &[
                                    (
                                        "/*",
                                        "*/",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Hex =>
                    Language::new_blank()
                ,Html =>
                    Language::new_html()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Idris =>
                    Language::new_haskell()
                
                    .nested()
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                    ]),IntelHex =>
                    Language::new_blank()
                ,Isabelle =>
                            Language::new(
                                &[
                                    "--",
                                ],
                                &[
                                    (
                                        "\\<open>",
                                        "\\<close>",
                                    ),
                                    (
                                        "‹",
                                        "›",
                                    ),
                                    (
                                        "{*",
                                        "*}",
                                    ),
                                    (
                                        "(*",
                                        "*)",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "''",
                                "''",
                            ),
                    ]),Jai =>
                    Language::new_c()
                
                    .nested(),Java =>
                    Language::new_c()
                ,JavaScript =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                            (
                                "`",
                                "`",
                            ),
                    ]),Json =>
                    Language::new_blank()
                ,Jsx =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                            (
                                "`",
                                "`",
                            ),
                    ]),Julia =>
                            Language::new(
                                &[
                                    "#",
                                ],
                                &[
                                    (
                                        "#=",
                                        "=#",
                                    ),
                                ]
                            )
                        
                    
                    .nested()
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Julius =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                            (
                                "`",
                                "`",
                            ),
                    ]),KakouneScript =>
                            Language::new_single(&[
                                    "#",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Kotlin =>
                    Language::new_c()
                
                    .nested()
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Lean =>
                            Language::new(
                                &[
                                    "--",
                                ],
                                &[
                                    (
                                        "/-",
                                        "-/",
                                    ),
                                ]
                            )
                        
                    
                    .nested(),Less =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),LinkerScript =>
                    Language::new_c()
                ,Lisp =>
                            Language::new(
                                &[
                                    ";",
                                ],
                                &[
                                    (
                                        "#|",
                                        "|#",
                                    ),
                                ]
                            )
                        
                    
                    .nested(),Lua =>
                            Language::new(
                                &[
                                    "--",
                                ],
                                &[
                                    (
                                        "--[[",
                                        "]]",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Lucius =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Madlang =>
                            Language::new(
                                &[
                                    "#",
                                ],
                                &[
                                    (
                                        "{#",
                                        "#}",
                                    ),
                                ]
                            )
                        
                    ,Makefile =>
                    Language::new_hash()
                ,Markdown =>
                    Language::new_blank()
                ,Meson =>
                            Language::new_single(&[
                                    "#",
                            ])
                    
                    .set_quotes(&[
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Mint =>
                    Language::new_blank()
                ,ModuleDef =>
                            Language::new_single(&[
                                    ";",
                            ])
                    ,MsBuild =>
                    Language::new_html()
                ,Mustache =>
                        Language::new_multi(&[
                                (
                                    "{{!",
                                    "}}",
                                ),
                        ])
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Nim =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Nix =>
                    Language::new_c()
                ,OCaml =>
                    Language::new_func()
                ,ObjectiveC =>
                    Language::new_c()
                ,ObjectiveCpp =>
                    Language::new_c()
                ,Org =>
                            Language::new_single(&[
                                    "# ",
                            ])
                    ,Oz =>
                    Language::new_pro()
                ,PSL =>
                    Language::new_c()
                ,Pascal =>
                        Language::new_multi(&[
                                (
                                    "(*",
                                    "*)",
                                ),
                                (
                                    "{",
                                    "*)",
                                ),
                                (
                                    "(*",
                                    "}",
                                ),
                                (
                                    "{",
                                    "}",
                                ),
                        ])
                    .set_quotes(&[
                            (
                                "'",
                                "'",
                            ),
                    ]),Perl =>
                            Language::new(
                                &[
                                    "#",
                                ],
                                &[
                                    (
                                        "=pod",
                                        "=cut",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Php =>
                            Language::new(
                                &[
                                    "#",
                                    "//",
                                ],
                                &[
                                    (
                                        "/*",
                                        "*/",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Polly =>
                    Language::new_html()
                ,Processing =>
                    Language::new_c()
                ,Prolog =>
                    Language::new_pro()
                ,Protobuf =>
                            Language::new_single(&[
                                    "//",
                            ])
                    ,PureScript =>
                    Language::new_haskell()
                ,Python =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Qcl =>
                    Language::new_c()
                ,Qml =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),R =>
                    Language::new_hash()
                ,Racket =>
                            Language::new(
                                &[
                                    ";",
                                ],
                                &[
                                    (
                                        "#|",
                                        "|#",
                                    ),
                                ]
                            )
                        
                    
                    .nested(),Rakefile =>
                            Language::new(
                                &[
                                    "#",
                                ],
                                &[
                                    (
                                        "=begin",
                                        "=end",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Razor =>
                        Language::new_multi(&[
                                (
                                    "<!--",
                                    "-->",
                                ),
                                (
                                    "@*",
                                    "*@",
                                ),
                        ]),ReStructuredText =>
                    Language::new_blank()
                ,Ruby =>
                            Language::new(
                                &[
                                    "#",
                                ],
                                &[
                                    (
                                        "=begin",
                                        "=end",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),RubyHtml =>
                    Language::new_html()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Rust =>
                    Language::new_c()
                
                    .nested()
                    .set_quotes(&[
                            (
                                "r#\"",
                                "\"#",
                            ),
                            (
                                "#\"",
                                "\"#",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                    ]),SRecode =>
                            Language::new_single(&[
                                    ";;",
                            ])
                    ,Sass =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Scala =>
                    Language::new_c()
                ,Scheme =>
                            Language::new(
                                &[
                                    ";",
                                ],
                                &[
                                    (
                                        "#|",
                                        "|#",
                                    ),
                                ]
                            )
                        
                    
                    .nested(),Scons =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Sh =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Sml =>
                    Language::new_func()
                ,SpecmanE =>
                            Language::new(
                                &[
                                    "--",
                                    "//",
                                ],
                                &[
                                    (
                                        "'>",
                                        "<'",
                                    ),
                                ]
                            )
                        
                    ,Spice =>
                            Language::new_single(&[
                                    "*",
                            ])
                    ,Sql =>
                            Language::new(
                                &[
                                    "--",
                                ],
                                &[
                                    (
                                        "/*",
                                        "*/",
                                    ),
                                ]
                            )
                        
                    
                    .set_quotes(&[
                            (
                                "'",
                                "'",
                            ),
                    ]),Svg =>
                    Language::new_html()
                ,Swift =>
                    Language::new_c()
                
                    .nested(),SystemVerilog =>
                    Language::new_c()
                ,Tcl =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Tex =>
                            Language::new_single(&[
                                    "%",
                            ])
                    ,Text =>
                    Language::new_blank()
                ,Toml =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"\"\"",
                                "\"\"\"",
                            ),
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),TypeScript =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                            (
                                "`",
                                "`",
                            ),
                    ]),UnrealScript =>
                    Language::new_c()
                ,UrWeb =>
                    Language::new_func()
                ,UrWebProject =>
                            Language::new_single(&[
                                    "#",
                            ])
                    ,Vala =>
                    Language::new_c()
                ,Verilog =>
                    Language::new_c()
                ,VerilogArgsFile =>
                        Language::new_multi(&[
                        ]),Vhdl =>
                            Language::new_single(&[
                                    "--",
                            ])
                    ,VimScript =>
                            Language::new_single(&[
                                    "\"",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),VisualBasic =>
                            Language::new_single(&[
                                    "'",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                    ]),Vue =>
                    Language::new_html()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                            (
                                "`",
                                "`",
                            ),
                    ]),Wolfram =>
                    Language::new_func()
                ,XSL =>
                    Language::new_html()
                ,Xaml =>
                    Language::new_html()
                ,Xml =>
                    Language::new_html()
                ,Xtend =>
                    Language::new_c()
                
                    .set_quotes(&[
                            (
                                "'''",
                                "'''",
                            ),
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Yaml =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),Zig =>
                            Language::new_single(&[
                                    "//",
                            ])
                    
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "\\",
                                "
",
                            ),
                    ]),Zsh =>
                    Language::new_hash()
                
                    .set_quotes(&[
                            (
                                "\"",
                                "\"",
                            ),
                            (
                                "'",
                                "'",
                            ),
                    ]),
        }
    }
}

impl From<String> for LanguageType {
    fn from(from: String) -> Self {
        LanguageType::from(&*from)
    }
}

impl<'a> From<&'a str> for LanguageType {
    fn from(from: &str) -> Self {
        match &*from {
                    "ABAP"
                
                    =>Abap,
                    "ActionScript"
                    =>ActionScript,
                    "Ada"
                    =>Ada,
                    "Agda"
                    =>Agda,
                    "Alex"
                    =>Alex,
                    "ASP"
                
                    =>Asp,
                    "ASP.NET"
                
                    =>AspNet,
                    "Assembly"
                    =>Assembly,
                    "AutoHotKey"
                    =>AutoHotKey,
                    "Autoconf"
                    =>Autoconf,
                    "BASH"
                
                    =>Bash,
                    "Batch"
                    =>Batch,
                    "C"
                    =>C,
                    "C Header"
                
                    =>CHeader,
                    "CMake"
                    =>CMake,
                    "C#"
                
                    =>CSharp,
                    "C Shell"
                
                    =>CShell,
                    "Cabal"
                    =>Cabal,
                    "Cassius"
                    =>Cassius,
                    "Ceylon"
                    =>Ceylon,
                    "Clojure"
                    =>Clojure,
                    "ClojureC"
                    =>ClojureC,
                    "ClojureScript"
                    =>ClojureScript,
                    "COBOL"
                
                    =>Cobol,
                    "CoffeeScript"
                    =>CoffeeScript,
                    "Cogent"
                    =>Cogent,
                    "ColdFusion"
                    =>ColdFusion,
                    "ColdFusion CFScript"
                
                    =>ColdFusionScript,
                    "Coq"
                    =>Coq,
                    "C++"
                
                    =>Cpp,
                    "C++ Header"
                
                    =>CppHeader,
                    "Crystal"
                    =>Crystal,
                    "CSS"
                
                    =>Css,
                    "D"
                    =>D,
                    "Dart"
                    =>Dart,
                    "Device Tree"
                
                    =>DeviceTree,
                    "Dockerfile"
                    =>Dockerfile,
                    "Emacs Lisp"
                
                    =>Elisp,
                    "Elixir"
                    =>Elixir,
                    "Elm"
                    =>Elm,
                    "Elvish"
                    =>Elvish,
                    "Emacs Dev Env"
                
                    =>EmacsDevEnv,
                    "Erlang"
                    =>Erlang,
                    "FEN"
                
                    =>FEN,
                    "F#"
                
                    =>FSharp,
                    "Fish"
                    =>Fish,
                    "Forth"
                    =>Forth,
                    "FORTRAN Legacy"
                
                    =>FortranLegacy,
                    "FORTRAN Modern"
                
                    =>FortranModern,
                    "F*"
                
                    =>Fstar,
                    "GDScript"
                
                    =>GdScript,
                    "GLSL"
                
                    =>Glsl,
                    "Go"
                    =>Go,
                    "Groovy"
                    =>Groovy,
                    "Hamlet"
                    =>Hamlet,
                    "Handlebars"
                    =>Handlebars,
                    "Happy"
                    =>Happy,
                    "Haskell"
                    =>Haskell,
                    "Haxe"
                    =>Haxe,
                    "HCL"
                
                    =>Hcl,
                    "HEX"
                
                    =>Hex,
                    "HTML"
                
                    =>Html,
                    "Idris"
                    =>Idris,
                    "Intel HEX"
                
                    =>IntelHex,
                    "Isabelle"
                    =>Isabelle,
                    "JAI"
                
                    =>Jai,
                    "Java"
                    =>Java,
                    "JavaScript"
                    =>JavaScript,
                    "JSON"
                
                    =>Json,
                    "JSX"
                
                    =>Jsx,
                    "Julia"
                    =>Julia,
                    "Julius"
                    =>Julius,
                    "Kakoune script"
                
                    =>KakouneScript,
                    "Kotlin"
                    =>Kotlin,
                    "Lean"
                    =>Lean,
                    "LESS"
                
                    =>Less,
                    "LD Script"
                
                    =>LinkerScript,
                    "Lisp"
                    =>Lisp,
                    "Lua"
                    =>Lua,
                    "Lucius"
                    =>Lucius,
                    "Madlang"
                    =>Madlang,
                    "Makefile"
                    =>Makefile,
                    "Markdown"
                    =>Markdown,
                    "Meson"
                    =>Meson,
                    "Mint"
                    =>Mint,
                    "Module-Definition"
                
                    =>ModuleDef,
                    "MSBuild"
                
                    =>MsBuild,
                    "Mustache"
                    =>Mustache,
                    "Nim"
                    =>Nim,
                    "Nix"
                    =>Nix,
                    "OCaml"
                    =>OCaml,
                    "Objective C"
                
                    =>ObjectiveC,
                    "Objective C++"
                
                    =>ObjectiveCpp,
                    "Org"
                    =>Org,
                    "Oz"
                    =>Oz,
                    "PSL Assertion"
                
                    =>PSL,
                    "Pascal"
                    =>Pascal,
                    "Perl"
                    =>Perl,
                    "PHP"
                
                    =>Php,
                    "Polly"
                    =>Polly,
                    "Processing"
                    =>Processing,
                    "Prolog"
                    =>Prolog,
                    "Protocol Buffers"
                
                    =>Protobuf,
                    "PureScript"
                    =>PureScript,
                    "Python"
                    =>Python,
                    "QCL"
                
                    =>Qcl,
                    "QML"
                
                    =>Qml,
                    "R"
                    =>R,
                    "Racket"
                    =>Racket,
                    "Rakefile"
                    =>Rakefile,
                    "Razor"
                    =>Razor,
                    "ReStructuredText"
                    =>ReStructuredText,
                    "Ruby"
                    =>Ruby,
                    "Ruby HTML"
                
                    =>RubyHtml,
                    "Rust"
                    =>Rust,
                    "SRecode Template"
                
                    =>SRecode,
                    "Sass"
                    =>Sass,
                    "Scala"
                    =>Scala,
                    "Scheme"
                    =>Scheme,
                    "Scons"
                    =>Scons,
                    "Shell"
                
                    =>Sh,
                    "Standard ML (SML)"
                
                    =>Sml,
                    "Specman e"
                
                    =>SpecmanE,
                    "Spice Netlist"
                
                    =>Spice,
                    "SQL"
                
                    =>Sql,
                    "SVG"
                
                    =>Svg,
                    "Swift"
                    =>Swift,
                    "SystemVerilog"
                    =>SystemVerilog,
                    "TCL"
                
                    =>Tcl,
                    "TeX"
                
                    =>Tex,
                    "Plain Text"
                
                    =>Text,
                    "TOML"
                
                    =>Toml,
                    "TypeScript"
                    =>TypeScript,
                    "Unreal Script"
                
                    =>UnrealScript,
                    "Ur/Web"
                
                    =>UrWeb,
                    "Ur/Web Project"
                
                    =>UrWebProject,
                    "Vala"
                    =>Vala,
                    "Verilog"
                    =>Verilog,
                    "Verilog Args File"
                
                    =>VerilogArgsFile,
                    "VHDL"
                
                    =>Vhdl,
                    "Vim Script"
                
                    =>VimScript,
                    "Visual Basic"
                
                    =>VisualBasic,
                    "Vue"
                
                    =>Vue,
                    "Wolfram"
                    =>Wolfram,
                    "XSL"
                
                    =>XSL,
                    "XAML"
                
                    =>Xaml,
                    "XML"
                
                    =>Xml,
                    "Xtend"
                    =>Xtend,
                    "YAML"
                
                    =>Yaml,
                    "Zig"
                    =>Zig,
                    "Zsh"
                    =>Zsh,
            _ => unreachable!(),
        }
    }
}

impl fmt::Display for LanguageType {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.name())
    }
}


impl<'a> From<LanguageType> for Cow<'a, LanguageType> {
    fn from(from: LanguageType) -> Self {
        Cow::Owned(from)
    }
}

impl<'a> From<&'a LanguageType> for Cow<'a, LanguageType> {
    fn from(from: &'a LanguageType) -> Self {
        Cow::Borrowed(from)
    }
}


/// This is for getting the file type from the first line of a file
pub fn get_filetype_from_shebang<P>(file: P) -> Option<&'static str>
    where P: AsRef<Path>
{
    let file = match File::open(file) {
        Ok(file) => file,
        _ => return None,
    };

    let mut buf = BufReader::new(file);
    let mut line = String::new();
    let _ = buf.read_line(&mut line);

    let mut words = line.split_whitespace();
    match words.next() {
        Some("#!/bin/sh") => Some("sh"),
        Some("#!/bin/csh") => Some("csh"),
        Some("#!/usr/bin/perl") => Some("pl"),
        Some("#!/usr/bin/env") => {
            if let Some(word) = words.next() {
                match word {
                                "bash"
                                => Some("bash"),
                                "csh"
                                => Some("csh"),
                                "elvish"
                                => Some("elv"),
                                "fish"
                                => Some("fish"),
                                "python"
                                    |
                                "python2"
                                    |
                                "python3"
                                => Some("py"),
                                "sh"
                                => Some("sh"),
                    env => {
                        warn!("Unknown environment: {:?}", env);
                        None
                    }
                }
            } else {
                None
            }
        }
        _ => None,
    }
}
